@@ -1,9 +1,9 @@
 summary.uncertainty <-
 function(object, ndigits = 3, ...) {
-	if (class(object) != "uncertainty") {
+	if (!inherits(object, "uncertainty")) {
 	  stop("object must be an uncertainty object.\n")
 	}
-	if (all(object$method != c("Kragten","GFO","GSO","MC"))) {
+	if (all(object$method != c("GFO","GSO","MC"))) {
 	  stop("object was created with an unknown method.\n")
 	}
 	res <- list(call = object$call, measurand_name = object$measurand_name,
