@@ -1,30 +1,35 @@
 .internal_build_envir <-
 function(name, value) {
   n <- length(name)
-  res <- paste(name[1], "=", value[1], sep = "")
+  res <- paste0(name[1], "=", value[1])
   if (n > 1) {
     for (i in 2:n) {
-      res <- paste(res, ", ", paste(name[i], "=", value[i], sep = ""), sep = "")
+      res <- paste0(res, ", ", name[i], "=", value[i])
     }
   }
-  res <- paste("list(", res, ")", sep = "")
+  res <- paste0("list(", res, ")")
   return(res)
 }
 .internal_dof_eff <-
-function(c_u, dof, cor = diag(1, length(c_u))) {
-	ct <- sqrt( t(c_u) %*% cor %*% c_u )
+function(c_u, dof, cor = diag(1, length(c_u)), use.correlation = FALSE) {
+	ct <- c(sqrt( t(c_u) %*% cor %*% c_u )) # it is a scalar, without c() it appeared as a 1x1 matrix
 	res <- 0
 	rr_1 <- sum(c_u ^ 4 / dof)
 	rr_2 <- 0
 	rr_3 <- 0
 	n <- length(c_u)
-	# this implements the Castrup.2010 algorithm
-	for (i in 1:(n - 1)) {
-		for (j in (i + 1):n) {
-			rr_2 <- rr_2 + cor[i, j] ^ 2 * (c_u[i] ^ 2 / dof[i]) *
-			  (c_u[j] ^ 2 / dof[j]) * (dof[i] + dof[j] + 1 / 2)
-			rr_3 <- rr_3 + cor[i, j] * (c_u[i] ^ 3 / dof[i] * c_u[j] +
-			                              c_u[i] * c_u[j] ^ 3 / dof[j])
+	# corrections if correlation is present
+	if (use.correlation & any(cor != diag(1, n))) {
+		# this implements the Castrup.2010 algorithm, this approximation works fine for 
+		# medium dof in c(15, 100) and only two input quantities.
+		# Additional terms are missing when 3 or more input quantities are present.
+		for (i in 1:(n - 1)) {
+			for (j in (i + 1):n) {
+				rr_2 <- rr_2 + cor[i, j] ^ 2 * (c_u[i] ^ 2) *
+				  (c_u[j] ^ 2) * (1/dof[i] + 1/dof[j] + 1 / (2 * dof[i] * dof[j]) )
+				rr_3 <- rr_3 + cor[i, j] * (c_u[i] ^ 3 / dof[i] * c_u[j] +
+				                              c_u[i] * c_u[j] ^ 3 / dof[j])
+			}
 		}
 	}
 	res <- ct ^ 4 / (rr_1 + rr_2 + 2 * rr_3)
@@ -33,7 +38,8 @@ function(c_u, dof, cor = diag(1, length(c_u))) {
 
 
 .internal_gfo <-
-function(name, label = name, model, ub, alpha = 0.05) {
+function(name, label = name, model, description, ub, alpha = 0.05, use.correlation = FALSE) {
+	if (missing(use.correlation) | is.null(use.correlation)) use.correlation = FALSE
 	envir_str <- .internal_build_envir(ub$name, ub$mean)
 	n <- length(ub$name)
 	
@@ -42,6 +48,8 @@ function(name, label = name, model, ub, alpha = 0.05) {
 	}
 
 	data_cov <- matrix(ub$u, n, n) * ub$cor * t(matrix(ub$u, n, n))
+	
+#	print( data_cov )
 
 	# get the partial derivatives of the expression wrt each of the variables
 	# compute the contribution to the uncertainty by evaluating the list of
@@ -60,6 +68,8 @@ function(name, label = name, model, ub, alpha = 0.05) {
 
 	res$contribution <- res$coefficient * ub$u
 
+#	print(res$contribution)
+
 	# correlation contributions
 	JJ <- res$coefficient %*% t(res$coefficient)
 
@@ -68,46 +78,79 @@ function(name, label = name, model, ub, alpha = 0.05) {
 
 	cor_contribution <- GUM_u ^ 2 - sum(res$contribution ^ 2)
 
-	GUM_dof <- .internal_dof_eff(c_u = res$contribution, dof = ub$dof,
-	                             cor = ub$cor)
+	GUM_dof <- .internal_dof_eff(c_u = res$contribution, dof = ub$dof, 
+			cor = ub$cor, use.correlation = use.correlation)
 
-#	print(GUM_dof)
-	
-	GUM_U <- qt(1 - alpha / 2, GUM_dof) * GUM_u
+	if (!is.na(GUM_dof)) {
+		GUM_U <- qt(1 - alpha / 2, GUM_dof) * GUM_u
+
+		su <- NA
+		if (GUM_dof > 2) {
+			su <- GUM_u * sqrt(GUM_dof / (GUM_dof - 2))
+		}
+
+		if (GUM_mu != 0) { Urelpct <- GUM_U/GUM_mu/100.0 } else Urelpct <- NA
+
+		result_GUM <- list(method = "GFO",
+			call = match.call(),
+			measurand_name = name,
+			measurand_model = model,
+			measurand_label = label,
+			measurand_description = description,
+			mean = GUM_mu,
+			sd = GUM_u,
+			u = GUM_u,
+			alpha = alpha,
+			dof = GUM_dof,
+			U = GUM_U,
+			Upct = Urelpct,
+			lcl = GUM_mu - GUM_U,
+			ucl = GUM_mu + GUM_U,
+
+			variables = list(name = ub$name, label = ub$label, mean = ub$mean, u = ub$u,
+			                 distribution = ub$distribution, dof = ub$dof,
+			                 coeff = res$coefficient, cor = ub$cor),
+			contribution = res$contribution,
+			cor_contribution = cor_contribution,
+
+			# additional ouput
+			partial = res$partial,
+			coeff = res$coefficient,
+			su = su)
+	} else {
+		warning("dof are NA, it was not possible to estimate the effective degrees of freedom.")	
+		result_GUM <- list(method = "GFO",
+			call = match.call(),
+			measurand_name = name,
+			measurand_model = model,
+			measurand_label = label,
+			measurand_description = description,
+			mean = GUM_mu,
+			sd = GUM_u,
+			u = GUM_u,
+			alpha = alpha,
+			dof = NA,
+			U = NA,
+			lcl = NA,
+			ucl = NA,
+
+			variables = list(name = ub$name, label = ub$label, mean = ub$mean, u = ub$u,
+			                 distribution = ub$distribution, dof = ub$dof,
+			                 coeff = res$coefficient, cor = ub$cor),
+			contribution = res$contribution,
+			cor_contribution = cor_contribution,
+
+			# additional ouput
+			partial = res$partial,
+			coeff = res$coefficient,
+			su = NA)
 
-	su <- NA
-	if (GUM_dof > 2) {
-		su <- GUM_u * sqrt(GUM_dof / (GUM_dof - 2))
 	}
-
-	result_GUM <- list(method = "GFO",
-		call = match.call(),
-		measurand_name = name,
-		measurand_model = model,
-		measurand_label = label,
-		mean = GUM_mu,
-		sd = GUM_u,
-		u = GUM_u,
-		alpha = alpha,
-		dof = GUM_dof,
-		U = GUM_U,
-		lcl = GUM_mu - GUM_U,
-		ucl = GUM_mu + GUM_U,
-
-		variables = list(name = ub$name, label = ub$label, mean = ub$mean, u = ub$u,
-		                 dof = ub$dof, coeff = res$coefficient, cor = ub$cor),
-		contribution = res$contribution,
-		cor_contribution = cor_contribution,
-
-		# additional ouput
-		partial = res$partial,
-		coeff = res$coefficient,
-		su = su)
-
 	return(result_GUM)
 }
 .internal_gso <-
-function(name, label = name, model, ub, alpha = 0.05) {
+function(name, label = name, model, description, ub, alpha = 0.05, use.correlation = FALSE) {
+	if (missing(use.correlation) | is.null(use.correlation)) use.correlation = FALSE
 	envir_str <- .internal_build_envir(ub$name, ub$mean)
 	n <- length(ub$name)
 
@@ -150,41 +193,85 @@ function(name, label = name, model, ub, alpha = 0.05) {
 	# second order variance
 	u_so <- sqrt(GFO_u ^ 2 + sum(diag(H %*% data_cov %*% H %*% data_cov)) / 2)
 
-	dof_so <- .internal_dof_eff(c_u = res$contribution, dof = ub$dof,
-	                            cor = ub$cor)
+	dof_so <- .internal_dof_eff(c_u = res$contribution, dof = ub$dof, 
+			cor = ub$cor, use.correlation = use.correlation)
 
-	cor_contribution <- u_so ^ 2 - sum(res$contribution ^ 2)
+	su <- NA
+	if (!is.na(dof_so)) {
+		if (dof_so > 2) {
+			su <- u_so * sqrt(dof_so / (dof_so - 2))
+		}
 
-	U_so <- qt(1 - alpha / 2, dof_so) * u_so
+		cor_contribution <- u_so ^ 2 - sum(res$contribution ^ 2)
 
-	res <- list(method = "GSO",
-		call = match.call(),
-		measurand_name = name,
-		measurand_model = model,
-		measurand_label = label,
-		mean = mu_so,
-		sd = u_so,
-		u = u_so,
-		alpha = alpha,
-		dof = dof_so,
-		U = U_so,
-		lcl = mu_so - U_so,
-		ucl = mu_so + U_so,
+		U_so <- qt(1 - alpha / 2, dof_so) * u_so
 
-		variables = list(name = ub$name, label = ub$label, mean = ub$mean, u = ub$u,
-		                 dof = ub$dof, coeff = res$coefficient, cor = ub$cor),
-		contribution = res$contribution,
-		cor_contribution = cor_contribution,
+		if (mu_so != 0) { Urelpct <- U_so/mu_so/100.0 } else Urelpct <- NA
 
-		# additional ouput
-		partial = res$partial,
-		coeff = res$coefficient,
-		su = u_so * sqrt(dof_so / (dof_so - 2)))
+		res <- list(method = "GSO",
+			call = match.call(),
+			measurand_name = name,
+			measurand_model = model,
+			measurand_label = label,
+			measurand_description = description,
+			mean = mu_so,
+			sd = u_so,
+			u = u_so,
+			alpha = alpha,
+			dof = dof_so,
+			U = U_so,
+			Upct = Urelpct,
+			lcl = mu_so - U_so,
+			ucl = mu_so + U_so,
 
+			variables = list(name = ub$name, label = ub$label, mean = ub$mean, u = ub$u,
+		                 distribution = ub$distribution, dof = ub$dof,
+		                 coeff = res$coefficient, cor = ub$cor),
+
+					contribution = res$contribution,
+			cor_contribution = cor_contribution,
+
+			# additional ouput
+			partial = res$partial,
+			coeff = res$coefficient,
+			su = su
+		)
+	} else {
+		warning("dof are NA, it was not possible to estimate the effective degrees of freedom.")
+		cor_contribution <- u_so ^ 2 - sum(res$contribution ^ 2)
+
+		res <- list(method = "GSO",
+			call = match.call(),
+			measurand_name = name,
+			measurand_model = model,
+			measurand_label = label,
+			measurand_description = description,
+			mean = mu_so,
+			sd = u_so,
+			u = u_so,
+			alpha = alpha,
+			dof = NA,
+			U = NA,
+			lcl = NA,
+			ucl = NA,
+
+			variables = list(name = ub$name, label = ub$label, mean = ub$mean, u = ub$u,
+			                 distribution = ub$distribution, dof = ub$dof,
+			                 coeff = res$coefficient, cor = ub$cor),
+						contribution = res$contribution,
+						cor_contribution = cor_contribution,
+
+			# additional ouput
+			partial = res$partial,
+			coeff = res$coefficient,
+			su = su
+		)
+	}
 	return(res)
 }
 .internal_kragten <-
-function(name, label = name, model, ub, alpha = 0.05) {
+function(name, label = name, model, description, ub, alpha = 0.05, use.correlation = FALSE) {
+	if (missing(use.correlation) | is.null(use.correlation)) use.correlation = FALSE
 	envir_str <- .internal_build_envir(ub$name, ub$mean)
 	n <- length(ub$name)
 
@@ -206,40 +293,85 @@ function(name, label = name, model, ub, alpha = 0.05) {
 	                        Kragten_contribution))
 	Kragten_cor_contribution <- Kragten_u ^ 2 - sum(Kragten_contribution ^ 2)
 	Kragten_dof <- .internal_dof_eff(c_u = Kragten_contribution, dof = ub$dof,
-	                                 cor = ub$cor)
-	Kragten_U <- qt(1 - alpha / 2, Kragten_dof) * Kragten_u
-	Kragten_coeff <- Kragten_contribution
-	Kragten_coeff[ub$u > 0] <- Kragten_contribution[ub$u > 0] / ub$u[ub$u > 0]
-	Kragten_coeff[ub$u == 0] <- NA
+				cor = ub$cor, use.correlation = use.correlation)
 
-	result_Kragten <- list(method = "Kragten",
-		call = match.call(),
+	if (!is.na(Kragten_dof)) {
+	  su <- NA
+	  if (Kragten_dof > 2) {
+	    su <- Kragten_u * sqrt(Kragten_dof / (Kragten_dof - 2))
+	  }
+	  
+	  Kragten_U <- qt(1 - alpha / 2, Kragten_dof) * Kragten_u
+	  Kragten_coeff <- Kragten_contribution
+	  Kragten_coeff[ub$u > 0] <- Kragten_contribution[ub$u > 0] / ub$u[ub$u > 0]
+	  Kragten_coeff[ub$u == 0] <- NA
+	  
+ 	  if (Kragten_mean != 0) { Urelpct <- Kragten_U/Kragten_mean/100.0 } else Urelpct <- NA
+
+	  result_Kragten <- list(method = "Kragten",
+	                         call = match.call(),
+	                         measurand_name = name,
+	                         measurand_model = model,
+	                         measurand_label = label,
+				 measurand_description = description,
+	                         mean = Kragten_mean,
+	                         sd = Kragten_u,
+	                         u = Kragten_u,
+	                         alpha = alpha,
+	                         dof = Kragten_dof,
+	                         U = Kragten_U,
+					 Upct = Urelpct,
+	                         lcl = Kragten_mean - Kragten_U,
+	                         ucl = Kragten_mean + Kragten_U,
+	                         
+	                         variables = list(name = ub$name, label = ub$label, mean = ub$mean, u = ub$u,
+	                                          distribution = ub$distribution, dof = ub$dof,
+	                                          coeff = Kragten_coeff, cor = ub$cor),
+	                         
+	                         contribution = Kragten_contribution,
+	                         cor_contribution = Kragten_cor_contribution,
+	                         
+	                         su = su
+	                         
+	                         # no additional output
+	  )
+	} else {
+	  Kragten_coeff <- Kragten_contribution
+  	Kragten_coeff[ub$u > 0] <- Kragten_contribution[ub$u > 0] / ub$u[ub$u > 0]
+	  Kragten_coeff[ub$u == 0] <- NA
+
+  	result_Kragten <- list(method = "Kragten",
+	  	call = match.call(),
 		measurand_name = name,
-		measurand_model = model,
-		measurand_label = label,
+  		measurand_model = model,
+	  	measurand_label = label,
+		measurand_description = description,
 		mean = Kragten_mean,
-		sd = Kragten_u,
-		u = Kragten_u,
-		alpha = alpha,
-		dof = Kragten_dof,
-		U = Kragten_U,
-		lcl = Kragten_mean - Kragten_U,
-		ucl = Kragten_mean + Kragten_U,
+  		sd = Kragten_u,
+	  	u = Kragten_u,
+		 alpha = alpha,
+  		dof = NA,
+	  	U = NA,
+		lcl = NA,
+  		ucl = NA,
 
-		variables = list(name = ub$name, label = ub$label, mean = ub$mean, u = ub$u,
-		                 dof = ub$dof, coeff = Kragten_coeff, cor = ub$cor),
-		contribution = Kragten_contribution,
-		cor_contribution = Kragten_cor_contribution,
+	  	variables = list(name = ub$name, label = ub$label, mean = ub$mean, u = ub$u,
+		                   distribution = ub$distribution, dof = ub$dof,
+		                   coeff = Kragten_coeff, cor = ub$cor),
 
-		su = Kragten_u * sqrt(Kragten_dof / (Kragten_dof - 2))
+  		contribution = Kragten_contribution,
+	  	cor_contribution = Kragten_cor_contribution,
 
-		# no additional output
+		  su = NA
+
+  		# no additional output
 		)
-
+  }
 	return(result_Kragten)
 }
 .internal_mc <-
-function(name, label = name, model, ub, alpha = 0.05, B = 2000) {
+function(name, label = name, model, description, ub, alpha = 0.05, B = 2000, use.correlation = FALSE) {
+	if (missing(use.correlation) | is.null(use.correlation)) use.correlation = FALSE
 
 	n <- length(ub$name)
 
@@ -407,22 +539,34 @@ function(name, label = name, model, ub, alpha = 0.05, B = 2000) {
 	res_U <- (f_mc[order(f_mc)[round(B * (1 - alpha / 2))]] -
 	            f_mc[order(f_mc)[round(B * (alpha / 2))]]) / 2
 
+	res_dof_ws <- res_u^4/sum(res_contribution^4/ub$dof)
+
+	MC_dof <- .internal_dof_eff(c_u = res_contribution, dof = ub$dof,
+				cor = ub$cor, use.correlation = use.correlation)
+
+	if (res_mean != 0) { Urelpct <- res_U/res_mean/100.0 } else Urelpct <- NA
+
 	result_MC <- list(method = "MC",
 		call = match.call(),
 		measurand_name = name,
 		measurand_model = model,
 		measurand_label = label,
+		measurand_description = description,
 		mean = res_mean,
 		sd = res_u,
 		u = res_u,
 		alpha = alpha,
 		dof = res_dof_ws,
+		mc_dof = MC_dof,
 		U = res_U,
+		Upct = Urelpct,
 		lcl = f_mc[order(f_mc)[round(B * (alpha / 2))]],
 		ucl = f_mc[order(f_mc)[round(B * (1 - alpha / 2))]],
 
 		variables = list(name = ub$name, label = ub$label, mean = ub$mean, u = ub$u,
-		                 dof = ub$dof, coeff = res_coeff, cor = ub$cor),
+		                 distribution = ub$distribution, dof = ub$dof,
+		                 coeff = res_coeff, cor = ub$cor),
+
 		contribution = res_contribution,
 		cor_contribution = res_cor_contribution,
 
@@ -434,3 +578,14 @@ function(name, label = name, model, ub, alpha = 0.05, B = 2000) {
 
 	return(result_MC)
 }
+
+.format_value <-
+function(x, n = 2){
+# takes x as a matrix with two columns
+# (mean, u)
+# u is computed with n significant digits and mean is formated accordingly
+ndigs <- n - ceiling(log(abs(x[, 2]), 10))
+res2 <- round(x[, 2], ndigs)
+res1 <- round(x[, 1], ndigs)
+return(cbind(mean = res1, u = res2))
+}
